from __future__ import annotations
from typing import Set, List, Dict, Tuple, Optional, Any
from enum import Enum
import enum
from copy import deepcopy


class Operation(Enum):
    Union = enum.auto()
    Intersection = enum.auto()
    Subtraction = enum.auto()  # difference?


class Terminal:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, obj):
        if isinstance(obj, Terminal):
            return obj.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


# TODO ask - only name equality?
class Nonterminal:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, obj):
        if isinstance(obj, Nonterminal):
            return obj.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


class State:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, obj):
        if isinstance(obj, State):
            return obj.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


class REG:
    Rules = Dict[Nonterminal, Set[Tuple[Terminal, Optional[Nonterminal]]]]

    def __init__(self, nonterminals: Set[Nonterminal],
                 terminals: Set[Terminal],
                 rules: Rules,
                 init: Nonterminal):
        self.nonterminals = nonterminals
        self.terminals = terminals
        self.rules = rules
        self.init = init


class DFA:
    Transition = Dict[Tuple[State, Terminal], State]

    def set_or_none(self, out : Optional[Set[Any]]) -> Set[Any]:
        return set() if out is None else out

    def dict_or_none(self, out : Optional[Dict[Any]]) -> Set[Any]:
        return dict() if out is None else out

    def __init__(self, states: Set[State] = None,
                 terminals: Set[Terminal] = None,
                 transition: Transition = None,
                 init: State = None,
                 final: Set[State] = None):
        self.states = self.set_or_none(states)
        self.terminals = self.set_or_none(terminals)
        self.transition = self.dict_or_none(transition)
        self.init = init
        self.final = self.set_or_none(final)

    def total(self) -> DFA:
        dfa = deepcopy(self)
        # total: Transition
        total = {}

        # TODO consider import itertools
        hell_id = 0
        hell = State(str(hell_id))
        while hell in dfa.states:
            hell_id += 1
            hell.name = str(hell_id)
        dfa.states.add(hell)

        for state in dfa.states:
            for terminal in dfa.terminals:
                if (state, terminal) in dfa.transition:
                    total[state, terminal] = dfa.transition[state, terminal]
                else:
                    total[state, terminal] = hell

        dfa.transition = total
        return dfa

    def is_total(self) -> bool:
        for state in self.states:
            for terminal in self.terminals:
                if (state, terminal) not in self.transition:
                    return False
        return True

    def accepts(self, word: str) -> bool:
        state = self.init
        for i in range(len(word)):
            terminal = Terminal(word[i])
            if (state, terminal) not in self.transition:
                return False
            state = self.transition[state, terminal]
        return state in self.final

    def complement(self) -> DFA:
        dfa = deepcopy(self)
        if not dfa.is_total():
            dfa.total()

        new_final = set()
        for state in dfa.states:
            if state not in dfa.final:
                new_final.add(state)

        dfa.final = new_final
        return dfa

    @staticmethod
    def composition(dfa_1: DFA, dfa_2: DFA, operation: Operation) -> DFA:

        # 2019 IB102 slides want both total
        if not dfa_1.is_total():
            dfa_1 = dfa_1.total()
        if not dfa_2.is_total():
            dfa_2 = dfa_2.total()

        dfa = DFA()

        # new terminals
        dfa.terminals = dfa_1.terminals.union(dfa_2.terminals)
        # ATTENTION: terminals may differ -> make total at the end?

        # new states
        for state_1 in dfa_1.states:
            for state_2 in dfa_2.states:
                # f"{state_1.name}_{state_2.name}" ?
                state = State(state_1.name + "_" + state_2.name)
                dfa.states.add(state)

                # new init
                if not dfa.init and state_1 == dfa_1.init \
                        and state_2 == dfa_2.init:
                    dfa.init = state

                # new final
                if state_1 in dfa_1.final or state_2 in dfa_2.final:

                    if operation == Operation.Union:
                        dfa.final.add(state)

                    elif operation == Operation.Intersection and \
                            state_1 in dfa_1.final and \
                            state_2 in dfa_2.final:
                        dfa.final.add(state)

                    elif operation == Operation.Subtraction and \
                            state_2 not in dfa_2.final:  # TODO is it this?
                        dfa.final.add(state)

                # new transition
                for terminal in dfa.terminals:
                    if (state_1, terminal) in dfa_1.transition and \
                            (state_2, terminal) in dfa_2.transition:
                        move_1 = dfa_1.transition[state_1, terminal]
                        move_2 = dfa_2.transition[state_2, terminal]
                        move = State(move_1.name + "_" + move_2.name)
                        dfa.transition[state, terminal] = move

        return dfa

    @staticmethod
    def union(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Operation.Union)

    @staticmethod
    def intersection(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Operation.Intersection)

    @staticmethod
    def subtraction(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Operation.Subtraction)

    def eliminate_unreachable(self):  # -> DFA:
        dfa = deepcopy(self)

        previous = set()
        actual = {self.init}

        while previous != actual:
            previous = deepcopy(actual)

            for state in previous:
                for terminal in dfa.terminals:
                    if (state, terminal) in self.transition:
                        actual.add(self.transition[state, terminal])

        dfa.states = actual
        dfa.final = dfa.final.intersection(actual)
        for (state, terminal) in self.transition:
            if state not in actual:
                del dfa.transition[state, terminal]

        return dfa

    def sort_terminals(self) -> List[Terminal]:
        to_sort = deepcopy(self.terminals)
        to_sort = list(to_sort)
        names = []
        for terminal in to_sort:
            names.append((terminal.name, terminal))
        names.sort()
        sorted_terminals = []
        for name, terminal in names:
            sorted_terminals.append(terminal)
        return sorted_terminals

    def minimize(self) -> DFA:
        dfa = deepcopy(self)
        dfa = dfa.eliminate_unreachable()
        dfa = dfa.total()

        terminals = self.sort_terminals()  # bc I want to iterate always in same order

        previous = []
        actual = dict()
        for state in dfa.states:
            if state in dfa.final:
                actual[state] = 1
            else:
                actual[state] = 0

        while previous != actual:
            previous = deepcopy(actual)
            patterns = dict()
            actual = dict()

            for state in dfa.states:
                pattern = [previous[state]]
                for terminal in terminals:
                    result = dfa.transition[state, terminal]
                    pattern.append(previous[result])
                pattern = tuple(pattern)

                if pattern in patterns:
                    patterns[pattern].add(state)
                else:
                    patterns[pattern] = {state}

            i = 0
            for states in patterns.values():
                for state in states:
                    actual[state] = i
                i += 1

        new_transition = {}
        dfa.states = set()
        dfa.init = None
        dfa.final = set()
        for state in actual:
            new_state = State("new_" + str(actual[state]))
            dfa.states.add(new_state)

            for terminal in terminals:
                if not (new_state, terminal) in new_transition:
                    originally = dfa.transition[state, terminal]
                    now = State("new_" + str(actual[originally]))
                    new_transition[new_state, terminal] = now

            if state == self.init:
                dfa.init = new_state
            if state in self.final:
                dfa.final.add(new_state)

        dfa.transition = new_transition
        return dfa

    def rename_state(self, rename: State, name: str):
        new_state = State(name)
        self.states.add(new_state)
        self.states.remove(rename)
        transition = deepcopy(self.transition)

        if rename == self.init:
            self.init = new_state
        if rename in self.final:
            self.final.add(new_state)
            self.final.remove(rename)

        for terminal in self.terminals:
            if (rename, terminal) in self.transition:
                transition[new_state, terminal] = self.transition[rename, terminal]
                del transition[rename, terminal]

        for state, terminal in transition:
            if transition[state, terminal] == rename:
                del transition[state, terminal]
                transition[state, terminal] = new_state

        self.transition = transition

    def canonize(self) -> DFA:
        # TODO total?

        dfa = deepcopy(self)
        remain = deepcopy(self.states)
        terminals = dfa.sort_terminals()

        i = 0
        queue = [dfa.init]
        while len(queue) > 0:
            actual = queue.pop(0)

            for terminal in terminals:
                state = dfa.transition[actual, terminal]
                if state in remain:
                    queue.append(state)
                    remain.remove(state)

            dfa.rename_state(actual, str(i))
            i += 1
        return dfa


class NFA:
    def __init__(self, states: Set[State],
                 terminals: Set[Terminal],
                 transition: Dict[Tuple[State, Terminal], Set[State]],
                 init: State,
                 final: Set[State]):
        self.states = states
        self.terminals = terminals
        self.transition = transition
        self.init = init
        self.final = final
